/* globals __DEV__, __webpack_public_path__, __webpack_get_script_filename__, __repack__ */
import EventEmitter from 'events'; // @ts-ignore

import { NativeModules } from 'react-native';
import { LoadEvent } from '../shared/LoadEvent';
const CACHE_KEY = `Repack.ChunkManager.Cache.${__DEV__ ? 'debug' : 'release'}`;

class ChunkManagerBackend {
  forceRemoteChunkResolution = false;
  eventEmitter = new EventEmitter();

  configure(config) {
    var _config$forceRemoteCh;

    this.storage = config.storage;
    this.forceRemoteChunkResolution = (_config$forceRemoteCh = config.forceRemoteChunkResolution) !== null && _config$forceRemoteCh !== void 0 ? _config$forceRemoteCh : false;
    this.resolveRemoteChunk = config.resolveRemoteChunk;

    __repack__.loadChunkCallback.push = ((parentPush, ...data) => {
      this.eventEmitter.emit('loaded', data[0]);
      return parentPush(...data);
    }).bind(null, __repack__.loadChunkCallback.push.bind(__repack__.loadChunkCallback));
  }

  async initCache() {
    if (!this.cache) {
      var _await$this$storage$g, _this$storage;

      const cache = JSON.parse((_await$this$storage$g = await ((_this$storage = this.storage) === null || _this$storage === void 0 ? void 0 : _this$storage.getItem(CACHE_KEY))) !== null && _await$this$storage$g !== void 0 ? _await$this$storage$g : 'null');
      this.cache = {
        urls: {},
        ...cache
      };
    }
  }

  async saveCache() {
    var _this$storage2;

    await ((_this$storage2 = this.storage) === null || _this$storage2 === void 0 ? void 0 : _this$storage2.setItem(CACHE_KEY, JSON.stringify(this.cache)));
  }

  async resolveChunk(chunkId, parentChunkId) {
    var _global$__CHUNKS__, _global$__CHUNKS__$lo;

    await this.initCache();
    let fetch = false;
    let url;

    if (__DEV__ && !this.forceRemoteChunkResolution) {
      url = Chunk.fromDevServer(chunkId);
    } else if ((_global$__CHUNKS__ = global.__CHUNKS__) !== null && _global$__CHUNKS__ !== void 0 && (_global$__CHUNKS__$lo = _global$__CHUNKS__['local']) !== null && _global$__CHUNKS__$lo !== void 0 && _global$__CHUNKS__$lo.includes(chunkId) && !this.forceRemoteChunkResolution) {
      url = Chunk.fromFileSystem(chunkId);
    } else {
      if (!this.resolveRemoteChunk) {
        throw new Error('No remote chunk resolver was provided. Did you forget to add `ChunkManager.configure({ resolveRemoteChunk: ... })`?');
      }

      const location = await this.resolveRemoteChunk(chunkId, parentChunkId);
      url = Chunk.fromRemote(location.url, {
        excludeExtension: location.excludeExtension
      });
    }

    if (!this.cache.urls[chunkId] || this.cache.urls[chunkId] !== url) {
      fetch = true;
      this.cache.urls[chunkId] = url;
      await this.saveCache();
    }

    return {
      url: this.cache.urls[chunkId],
      fetch
    };
  }

  async loadChunk(chunkId, parentChunkId) {
    let url;
    let fetch;

    try {
      const resolved = await this.resolveChunk(chunkId, parentChunkId);
      url = resolved.url;
      fetch = resolved.fetch;
    } catch (error) {
      console.error('ChunkManager.resolveChunk error:', error.message);
      throw new LoadEvent('resolution', chunkId, error);
    }

    try {
      const loadedPromise = new Promise(resolve => {
        this.eventEmitter.once('loaded', data => {
          if (data === chunkId) {
            resolve();
          }
        });
      });
      await NativeModules.ChunkManager.loadChunk(chunkId, url, fetch);
      await loadedPromise;
    } catch (error) {
      const {
        message,
        code
      } = error;
      console.error('ChunkManager.loadChunk invocation failed:', message, code ? `[${code}]` : '');
      throw new LoadEvent('load', url, error);
    }
  }

  async preloadChunk(chunkId) {
    let url;
    let fetch;

    try {
      const resolved = await this.resolveChunk(chunkId);
      url = resolved.url;
      fetch = resolved.fetch;
    } catch (error) {
      console.error('ChunkManager.resolveChunk error:', error.message);
      throw new LoadEvent('resolution', chunkId, error);
    }

    try {
      await NativeModules.ChunkManager.preloadChunk(chunkId, url, fetch);
    } catch (error) {
      const {
        message,
        code
      } = error;
      console.error('ChunkManager.preloadChunk invocation failed:', message, code ? `[${code}]` : '');
      throw new LoadEvent('load', url, error);
    }
  }

  async invalidateChunks(chunksIds = []) {
    try {
      await this.initCache();
      const ids = chunksIds !== null && chunksIds !== void 0 ? chunksIds : Object.keys(this.cache.urls);

      for (const chunkId of ids) {
        var _this$cache;

        (_this$cache = this.cache) === null || _this$cache === void 0 ? true : delete _this$cache.urls[chunkId];
      }

      await this.saveCache();
      await NativeModules.ChunkManager.invalidateChunks(ids);
    } catch (error) {
      const {
        message,
        code
      } = error;
      console.error('ChunkManager.invalidateChunks invocation failed:', message, code ? `[${code}]` : '');
      throw error;
    }
  }

}
/**
 * A manager to ease resolving, downloading and executing additional code from async chunks or
 * any arbitrary JavaScript files.
 *
 * - In development mode, all chunks will be resolved and downloaded from the Development server.
 * - In production mode, local chunks will be resolved and loaded from filesystem and remote
 * chunks will be resolved and downloaded based on the `resolveRemoteChunk` function.
 * - You can force all resolution, regardless of the mode, to go through `resolveRemoteChunk`
 * function by setting `forceRemoteChunkResolution: true` in `ChunkManager.configure(...)`.
 *
 * This API is only useful if you are working with any form of Code Splitting.
 *
 * Example of using this API with async chunks:
 * ```js
 * import * as React from 'react';
 * import { ChunkManager } from '@callstack/repack/client';
 *
 * ChunkManager.configure({
 *   resolveRemoteChunk: async (chunkId) => {
 *     return {
 *       url: `http://domain.exaple/apps/${chunkId}`,
 *     };
 *   },
 * });
 *
 * // ChunkManager.loadChunk is called internally when running `import()`
 * const TeacherModule = React.lazy(() => import('./Teacher.js'));
 * const StudentModule = React.lazy(() => import('./Student.js'));
 *
 * export function App({ role }) {
 *   if (role === 'teacher') {
 *     return <TeacherModule />;
 *   }
 *
 *   return <StudentModule />
 * }
 * ```
 */


export class ChunkManager {
  /**
   * A instance of `ChunkManagerBackend`.
   * Should not be used directly.
   *
   * @internal
   */
  static backend = new ChunkManagerBackend();
  /**
   * Configures `ChunkManager` to be able to resolve location of additional
   * chunks (or arbitrary code) in production.
   * Optionally, it also allows to set up caching to avoid over-fetching of chunks.
   *
   * @param config Configuration options.
   */

  static configure(config) {
    ChunkManager.backend.configure(config);
  }
  /**
   * Resolves a URL to a given chunks and  whether to download a chunk
   * or reuse previously downloaded one.
   *
   * @param chunkId Id of the chunk.
   * @returns Promise with chunk's URL as `url` and a boolean `fetch` whether to download a chunk
   * or reuse previously downloaded one.
   */


  static async resolveChunk(chunkId) {
    return ChunkManager.backend.resolveChunk(chunkId);
  }
  /**
   * Resolves given chunk's location, download and execute it.
   *
   * This function can be awaited for error handling, but after the functions resolves,
   * it should **not be assumed that the code was executed**.
   *
   * The execution of the code is handled internally by threading in React Native.
   *
   * @param chunkId Id of the chunk.
   * @param parentChunkId Id of the parent chunk.
   */


  static async loadChunk(chunkId, parentChunkId) {
    return ChunkManager.backend.loadChunk(chunkId, parentChunkId);
  }
  /**
   * Resolves given chunk's location and download it without executing.
   *
   * This function can be awaited to detect if the chunk was downloaded and for error handling.
   *
   * @param chunkId Id of the chunk.
   */


  static async preloadChunk(chunkId) {
    return ChunkManager.backend.preloadChunk(chunkId);
  }
  /**
   * Clears the cache (if configured in {@link ChunkManager.configure}) and removes downloaded
   * files for given chunks from the filesystem.
   *
   * This function can be awaited to detect if the chunks were invalidated and for error handling.
   *
   * @param chunksIds Array of chunk Ids to clear from cache and remove from filesystem.
   */


  static async invalidateChunks(chunksIds = []) {
    return ChunkManager.backend.invalidateChunks(chunksIds);
  }

}
/**
 * A helper class to ease the creation of of chunk based on it's location.
 *
 * **You should not need to use this.**
 *
 * @internal
 */

export class Chunk {
  /**
   * Creates definition for a chunk hosted on development server.
   *
   * @param chunkId Id of the chunk.
   * @returns Chunk definition.
   */
  static fromDevServer(chunkId) {
    return `${__webpack_public_path__}${__webpack_get_script_filename__(chunkId)}`;
  }
  /**
   * Creates definition for a chunk stored on filesystem on the target mobile device.
   *
   * @param chunkId Id of the chunk.
   * @returns Chunk definition.
   */


  static fromFileSystem(chunkId) {
    return __webpack_get_script_filename__(`file:///${chunkId}`);
  }
  /**
   * Creates definition for a chunk hosted on a remote server.
   *
   * @param url A URL to remote location where the chunk is stored.
   * @param options Additional options.
   * @returns Chunk definition.
   */


  static fromRemote(url, options = {}) {
    if (options.excludeExtension) {
      return url;
    }

    return __webpack_get_script_filename__(url);
  }

}
//# sourceMappingURL=ChunkManager.js.map